/*
 * Copyright (c) 2007 innoSysTec (R) GmbH, Germany. All rights reserved.
 * Original author: Edmund Wagner
 * Creation date: 22.05.2007
 *
 * Source: $HeadURL$
 * Last changed: $LastChangedDate$
 * 
 * the unrar licence applies to all junrar source and binary distributions 
 * you are not allowed to use this source to re-create the RAR compression algorithm
 * 
 * Here some html entities which can be used for escaping javadoc tags:
 * "&":  "&#038;" or "&amp;"
 * "<":  "&#060;" or "&lt;"
 * ">":  "&#062;" or "&gt;"
 * "@":  "&#064;" 
 */
package de.innosystec.unrar;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.innosystec.unrar.exception.RarException;
import de.innosystec.unrar.exception.RarException.RarExceptionType;
import de.innosystec.unrar.io.IReadOnlyAccess;
import de.innosystec.unrar.io.ReadOnlyAccessByteArray;
import de.innosystec.unrar.io.ReadOnlyAccessInputStream;
import de.innosystec.unrar.rarfile.AVHeader;
import de.innosystec.unrar.rarfile.BaseBlock;
import de.innosystec.unrar.rarfile.BlockHeader;
import de.innosystec.unrar.rarfile.CommentHeader;
import de.innosystec.unrar.rarfile.EndArcHeader;
import de.innosystec.unrar.rarfile.FileHeader;
import de.innosystec.unrar.rarfile.MainHeader;
import de.innosystec.unrar.rarfile.MarkHeader;
import de.innosystec.unrar.rarfile.ProtectHeader;
import de.innosystec.unrar.rarfile.SignHeader;
import de.innosystec.unrar.rarfile.UnrarHeadertype;
import de.innosystec.unrar.unpack.ComprDataIO;
import de.innosystec.unrar.unpack.Unpack;

/**
 * DOCUMENT ME
 *
 * @author $LastChangedBy$
 * @version $LastChangedRevision$
 */
public class Archive {
	
	private static Log logger = LogFactory.getLog(Archive.class.getName());
	
	private List<BaseBlock> headers = new ArrayList<BaseBlock>();
	
	private IReadOnlyAccess file;

	private ComprDataIO dataIO = new ComprDataIO(this);
	
	private MainHeader mainHeader = null;

	private Unpack unpack;

	
	/**
	 * create a new archive object using the given file
	 * @param file the file to extract
	 * @throws RarException 
	 */
	public Archive(IReadOnlyAccess file) throws RarException {
		this.file = file;
		try{
			readHeaders();
		}catch (Exception e) {
			logger.warn("exception in archive constructor maybe file is encrypted or currupt",e);
			//ignore exceptions to allow exraction of working files in corrupt archive
		}

	}
	
	/**
	 * 
	 * @return whether the archive is encrypted
	 */
	public boolean isEncrypted()
	{
		if(mainHeader!=null){
		return mainHeader.isEncrypted();
		}else{
			throw new NullPointerException("mainheader is null");
		}
	}
	
	/**
	 * Read the headers of the archive
	 * @throws RarException
	 */
	private void readHeaders() throws RarException{
		int toRead = 0;
		
		while(true){
			int size = 0;
			long newpos = 0;
			byte[] baseBlockBuffer =new byte[BaseBlock.BaseBlockSize];
			
			long position = file.getPosition();
			logger.info("\n--------reading header--------");
			size = file.readBytes(baseBlockBuffer, BaseBlock.BaseBlockSize);
			if (size == 0){
				break;
			}
			BaseBlock block = new BaseBlock(baseBlockBuffer);
			
			block.setPositionInFile(position);
			
			
			switch(block.getHeaderType())
			{
			
			case MarkHeader:
				MarkHeader markHead = new MarkHeader(block);
				headers.add(markHead);
				markHead.print();
				break;
			
			case MainHeader:
				int mainHeaderSize = 0;
				toRead = block.hasEncryptVersion() ? MainHeader.mainHeaderSizeWithEnc : MainHeader.mainHeaderSize;
				byte[] mainbuff = new byte[toRead];
				mainHeaderSize = file.readBytes(mainbuff, toRead);
				MainHeader mainhead =new MainHeader(block,mainbuff);
				headers.add(mainhead);
				this.mainHeader = mainhead;
				if(mainHeader.isEncrypted()){
					throw new RarException(RarExceptionType.rarEncryptedException);
				}
				mainhead.print();
				break;
			
			case SignHeader:
				int signHeaderSize = 0;
				toRead = SignHeader.signHeaderSize;
				byte[] signBuff = new byte[toRead];
				signHeaderSize = file.readBytes(signBuff, toRead);
				SignHeader signHead = new SignHeader(block,signBuff);
				headers.add(signHead);
				logger.info("HeaderType: SignHeader");
				
				break;
			
			case AvHeader:
				int avHeaderSize = 0;
				toRead = AVHeader.avHeaderSize;
				byte[] avBuff = new byte[toRead];
				avHeaderSize = file.readBytes(avBuff, toRead);
				AVHeader avHead = new AVHeader(block,avBuff);
				headers.add(avHead);
				logger.info("headertype: AVHeader");
				break;
			
			case CommHeader:
				int commHeaderSize = 0;
				toRead = CommentHeader.commentHeaderSize;
				byte[] commBuff = new byte[toRead];
				commHeaderSize = file.readBytes(commBuff, toRead);
				CommentHeader commHead = new CommentHeader(block,commBuff);
				headers.add(commHead);
				logger.info("method: "+commHead.getUnpMethod()+"; 0x"+Integer.toHexString(commHead.getUnpMethod()));
				newpos = commHead.getPositionInFile() + commHead.getHeaderSize();
				file.setPosition(newpos);
				
				break;
			case EndArcHeader:
				
				toRead = 0;
				toRead += block.hasArchiveDataCRC() ? EndArcHeader.endArcArchiveDataCrcSize : 0;
				toRead += block.hasVolumeNumber() ? EndArcHeader.endArcVolumeNumberSize : 0;
				EndArcHeader endArcHead;
				if(toRead > 0){
					int endArcHeaderSize = 0;
					byte[] endArchBuff = new byte[toRead];
					endArcHeaderSize = file.readBytes(endArchBuff, toRead);
					endArcHead = new EndArcHeader(block,endArchBuff);
					logger.info("HeaderType: endarch\ndatacrc:"+endArcHead.getArchiveDataCRC());
				}else{
					logger.info("HeaderType: endarch - no Data");
					endArcHead = new EndArcHeader(block,null);
				}
				headers.add(endArcHead);
				return;

			default:
				byte[] blockHeaderBuffer = new byte[BlockHeader.blockHeaderSize];
				int bhsize = file.readBytes(blockHeaderBuffer, BlockHeader.blockHeaderSize);
				BlockHeader blockHead = new BlockHeader(block,blockHeaderBuffer);
				
				switch(blockHead.getHeaderType())
				{
				case NewSubHeader:
				case FileHeader:
					toRead = blockHead.getHeaderSize()-BlockHeader.BaseBlockSize-BlockHeader.blockHeaderSize;
					byte[] fileHeaderBuffer = new byte[toRead];
					int fhsize = file.readBytes(fileHeaderBuffer, toRead);
					
					FileHeader fh = new FileHeader(blockHead,fileHeaderBuffer);
					
					fh.print();
					
					headers.add(fh);
					newpos = fh.getPositionInFile() + fh.getHeaderSize() + fh.getFullPackSize();
					file.setPosition(newpos);
					break;
				
				case ProtectHeader:
					toRead = blockHead.getHeaderSize()-BlockHeader.BaseBlockSize-BlockHeader.blockHeaderSize;
					byte[] protectHeaderBuffer = new byte[toRead];
					int phsize = file.readBytes(protectHeaderBuffer, toRead);
					ProtectHeader ph = new ProtectHeader(blockHead,protectHeaderBuffer);
					
					logger.info("totalblocks"+ph.getTotalBlocks());
					newpos = ph.getPositionInFile() + ph.getHeaderSize();
					file.setPosition(newpos);
					break;
				
				default:
					logger.warn("Unknown Header");
					throw new RarException(RarExceptionType.notRarArchive); 
					
				}
				
			}
		}
	}
	

	/**
	 * @return returns all headers of the archive
	 */
	public List<BaseBlock> getHeaders() {
		return headers;
	}

	/**
	 * @return returns all file headers of the archive
	 */
	public List<FileHeader> getFileHeaders(){
		List<FileHeader> list = new ArrayList<FileHeader>();
		for (BaseBlock block: headers) {
			if(block.getHeaderType().equals(UnrarHeadertype.FileHeader)){
				list.add((FileHeader)block);
			}
		}
		return list;
	}
	
	/**
	 * Extract the file specified by the given header and write it to the supplied output stream
	 * @param header the header to be extracted
	 * @param os the outputstream 
	 * @throws RarException
	 */
	public void extractFile(BaseBlock header,OutputStream os) throws RarException{
		FileHeader hd = null;
		if(header instanceof FileHeader){
			hd = (FileHeader) header;
			if(getHeaders().contains(hd)){
				
				try {
					doExtractFile(hd, os);
				} catch (Exception e) {
					if(e instanceof RarException){
						throw new RarException((RarException)e);
					}else{
						throw new RarException(e);
					}
				}
				
			}else{
				throw new RarException(RarExceptionType.headerNotInArchive);
			}
		}else{
			throw new RarException(RarExceptionType.wrongHeaderType);
		}
	}
	
	private void doExtractFile(FileHeader hd, OutputStream os) throws RarException
	{
		
		ReadOnlyAccessInputStream is = new ReadOnlyAccessInputStream(	
					file,
					hd.getPositionInFile()+hd.getHeaderSize(),
					hd.getPositionInFile()+hd.getHeaderSize()+hd.getFullPackSize()
					);
		
		
		dataIO.init(is, os);
		dataIO.setSubHeader(hd);
		dataIO.setPackedSizeToRead(hd.getFullPackSize());
		dataIO.setCurUnpRead(0);
		dataIO.setCurPackWrite(0);
		dataIO.setUnpFileCRC(this.isOldFormat()?0:0xffFFffFF);
		dataIO.setPackedCRC(0xFFffFFff);
		if(unpack==null){
			unpack = new Unpack(dataIO);
			
		}
		if(!hd.isSolid()){
			unpack.init(null);
		}
		unpack.setDestSize(hd.getFullUnpackSize());
		try {
			unpack.doUnpack(hd.getUnpVersion(), hd.isSolid());
			long hdfilecrc = hd.getFileCRC();
			long unpfilecrc = dataIO.getUnpFileCRC();
			logger.info("CRC "+(~dataIO.getUnpFileCRC()!=hd.getFileCRC()?"Failed\n":"OK\n")+"stored crc: "+Long.toHexString(hdfilecrc)+" :"+hdfilecrc+" current crc: \t"+Long.toHexString(~(unpfilecrc))+" : "+~unpfilecrc);
			if(~dataIO.getUnpFileCRC()!=hd.getFileCRC()){
				throw new RarException(RarExceptionType.crcError);
			}
		} catch (Exception e) {
			if(e instanceof RarException){
				throw new RarException((RarException)e);
			}else{
				throw new RarException(e);
			}
		}
		
		
	}

	
	/**
	 * @return returns the main header of this archive
	 */
	public MainHeader getMainHeader()
	{
		return mainHeader;
	}
	
	/**
	 * not yet supported
	 * @return whether the archive is old format
	 */
	public boolean isOldFormat()
	{
		// TODO Auto-generated method stub
		return false;
	}
	
	/**
	 * @param args
	 */
	public static void main(String[] args)
	{
				
		try {
			File file = null;
			if(args.length>0){
				logger.info("Trying to read archive: "+args[0]);
				file = new File(args[0]);
			}else{
				StringBuffer str = new StringBuffer();
				//bin files
				//str.append("C:\\winrar-test\\bin-x86\\");
				//str.append("server.rar");												// >ok<
				//str.append("asdf.rar");												// >ok<
				//str.append("visio.rar");												// >ok<
				
				//bin 64bit
				//str.append("C:\\winrar-test\\bin-x86_64\\");
				//str.append("server.rar");												// >ok<
				//str.append("asdf.rar");												// >ok<
				//str.append("visio.rar");												// >ok<
				
				
				//ppmd	
				//str.append("/home/edmund/Desktop/unrar/");
				//str.append("C:\\winrar-test\\ppmd\\");
				
				//str.append("server10.64.rar"); //modelorder 10 max 64mb				// >failed<
				//str.append("asdf4.rar"); //modelorder 2 max 2mb						// >failed<	

				//str.append("asdf6.rar"); //modelorder 2 max 2mb						// >ok<	
				//str.append("asdf7.rar"); //modelorder 2 max 2mb						// >ok<	
				//str.append("asdf8.rar"); //modelorder 2 max 2mb						// >ok<	
				
				//Delta
				//str.append("C:\\winrar-test\\delta\\");
				//str.append("server.rar"); //1 channel									// >ok<
				//str.append("server4c.rar"); //4 channels								// >ok<
				
				
				//Audio
				//str.append("C:\\winrar-test\\audio\\");
				//str.append("muh.rar"); 												// >ok< 
				//str.append("server.rar");//1 channel									// >ok<
				//str.append("server4c.rar");//4 channels								// >ok<
				//str.append("asdf5.rar");												// >ok<
				
				//True Color
				//str.append("C:\\winrar-test\\trueColor\\");
				//str.append("server.rar");												// >ok<
				//str.append("bmp.rar");												// >ok<
				
				
				//normal
				//str.append("C:\\winrar-test\\normal\\");
				//str.append("server.rar");												// >ok<
				//str.append("asdf.rar");												// >ok<
				//str.append("asdf3.rar");												// >ok<
				
				//solid
				//str.append("C:\\winrar-test\\solid\\");
				//str.append("solid2.rar");												// >ok<
				
				//multi volume
				//str.append("C:\\winrar-test\\multivol\\");
				//str.append("server.part01.rar");										// >ok but only the first part / file not complete<
				//str.append("server.part02.rar");										// >failed<
				
				//save only
				//str.append("C:\\winrar-test\\save\\");
				//str.append("server.rar");												// >ok<

				//multiple files not solid
				//str.append("C:\\winrar-test\\multiplefiles\\");
				//str.append("multifiles.rar");											// >ok<

				
				//encrypted
				str.append("C:\\winrar-test\\enc\\");
				//str.append("enc.rar");												// >failed< but recognized
				str.append("enc2.rar");												// >failed< but recognized
				//create file obj
				
				
				//bytearraytest
				//str.append("C:\\winrar-test\\");
				//str.append("brokenRar.rar");
				//str.append("blabla2.rar");											// >ok<
				//str.append("eclipse.rar");											// >ok<
				file = new File(str.toString());
				logger.info("trying to extract file: "+str.toString());
			}
			
			// create a readonlyfile object; this can be either a file or a byte array 
						
			ByteArrayOutputStream tempFile = new ByteArrayOutputStream();
			BufferedInputStream tempStream = new BufferedInputStream(new FileInputStream(file));
			while(true){
				try {
					int ch = tempStream.read();
					if(ch != -1){
						tempFile.write(ch);
					}else{
						break;
					}
				} catch (IOException e) {
					break;
				}
			}
//			ReadOnlyAccessFile f = new ReadOnlyAccessFile(file);
			ReadOnlyAccessByteArray f = new ReadOnlyAccessByteArray(tempFile.toByteArray());
			
			
			// try to read the archive
			Archive arc = null;
			try {
				arc = new Archive(f);
			} catch (RarException e) {
				logger.warn("archive consturctor error",e);
			}			
			
			arc.getFileHeaders().size();
			
			if(arc != null){
				if(arc.isEncrypted()){
					logger.warn("archive is encrypted cannot extreact");
					
				}
				
				// optional get all headers from archive later this can be used to get archive comments, recovery blocks etc
				List<BaseBlock> l = arc.getHeaders();
				
//				for (int i = 0; i < l.size(); i++) {
//					logger.info(l.get(i).getHeaderType());
//				}
				
				// get the file header from the archive this can be used to list the available files or to apply a file filter
				List<FileHeader> files = arc.getFileHeaders();
				
				// in this case we try to extract all files
				for(FileHeader fh : files)
				{
					if(fh.isEncrypted()){
						logger.warn("file is encrypted cannot extract: "+fh.getFileNameString());
						continue;
					}
					logger.info("extracting file:"+new String(fh.getFileNameByteArray()));
					logger.info("start: "+new Date());
					
					// create an output stream this can be what-so-ever output stream 
					// the file name can be aquired form the header
//					FileOutputStream os = new FileOutputStream(new String(fh.getFileNameByteArray()));
					ByteArrayOutputStream os = new ByteArrayOutputStream();
					
					
					try {
						arc.extractFile(fh, os);
					} catch (RarException e) {
						// TODO Auto-generated catch block
						logger.warn("error extracting file",e);
					}
					
					// if you're using byte arrays get the array from the stream 
					byte[] outMem = os.toByteArray();
					FileOutputStream fo = new FileOutputStream(new String(fh.getFileNameByteArray()));
					try {
						fo.write(outMem);
					} catch (IOException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					logger.info("end: "+new Date());
				}
			}
			
		} catch (FileNotFoundException e) {
			logger.error("File not found!!!");
		}
		
	}


	
	
	
	


}















